'use strict'
let txtdiv = document.createElement("div")
document.body.appendChild(txtdiv)
const output = {
  print: (string) => txtdiv.textContent += string
}
let wasmMem
let forth

/* Input capture */
let stdin = ""
document.addEventListener('keydown', (event) => {
  console.log(`keydown: ${event.key}`)
  if (event.key != "F5") {
    event.preventDefault()
    event.stopPropagation()
    switch (event.key) {
    case "Enter":
      txtdiv = document.createElement("div")
      document.body.appendChild(txtdiv)
      forth()
      output.print("__ok.")
      txtdiv = document.createElement("div")
      document.body.appendChild(txtdiv)
      break
    case "Backspace":
      stdin = stdin.substring(0, stdin.length - 1)
      txtdiv.textContent = txtdiv.textContent.substring(0, txtdiv.textContent.length - 1)
      break
    default:
      if (event.key.length == 1) {
	stdin += event.key
        output.print(event.key)
      }
      break
    }
  }
})

const channels = [{
  read: (writeAddr, maxBytes) => {
    const maxChars = maxBytes >> 1
    const bufView = new Uint16Array(wasmMem.buffer, writeAddr, maxChars)
    let i
    for (i = 0; i < maxChars && i < stdin.length; i++)
      bufView[i] = stdin.charCodeAt(i)
    stdin = stdin.substring(maxChars)
    return i << 1
  },
  write: (readAddr, maxBytes) =>
    output.print(String.fromCharCode.apply(
      null,
      new Uint16Array(wasmMem.buffer, readAddr, maxBytes)
    ))
}]
const simstack = []
const rstack = []
const dictionary = {
  ';': 1,
  'LIT': 2,
  RINIT: 3,
  WORD: 696,
  KEY: 5,
  DUP: 6,
  '+': 7,
  'NOOP2': 8,
  '.': 9,
  '@': 10,
  '!': 11,
  EXECUTE: 12,
  NOOP: 13,
  'JZ:': 14,
  'JNZ:': 15,
  DROP: 16,
  'WS?': 17,
  'JMP:': 18,
  'WPUTC': 19,
  'WB0': 20,
  'FIND': 21,
  'NUMBER': 22,
  'W!LEN': 23,
  'J-1:': 24,
  'BYE': 25,
  'SWAP': 26,
  'WORDS': 27,
  'HERE': 28,
  'DEFINE': 29,
  ':': 900,
  'MODE': 384,
  'WBUF': 256,
  'EXECUTE-MODE': 800,
  'QUIT': 600,
  'INTERPRET': 512
}
const wasmImport = {
  env: {
    pop: () => simstack.pop(),
    push: (val) => simstack.push(val),
    rinit: () => rstack.length = 0,
    rpop: () => rstack.pop(),
    rpush: (val) => rstack.push(val),
    sys_write: (channel, fromBuffer) => {
      if (channels[channel] === undefined)
	return
      const maxBytes = new DataView(
	wasmMem.buffer,
	fromBuffer,
	4
      ).getUint32(0,true)
      console.log(`write ch:${channel} addr:${fromBuffer} len:${maxBytes}`)
      channels[channel].write(fromBuffer + 4, maxBytes)
    },
    sys_read: (channel, toBuffer) => {
      console.log(`read ch:${channel} buf:${toBuffer} current: ${stdin}`)
      const lenView = new DataView(wasmMem.buffer, toBuffer, 8)
      const maxBytes = lenView.getUint32(0,true)
      console.log(`read blen:${lenView.getUint32(0,true)} cstrlen:${lenView.getUint32(4,true)}`)
      /* If the channel is undefined, or if there isn't enough room in
       * toBuffer for even one character, then, if there is enough
       * room for an int write a zero, exit */
      if (channels[channel] === undefined || maxBytes < 6) {
	if (maxBytes >= 4)
	  lenView.setUint32(4,0,true)
      }
      const numBytes = channels[channel].read(toBuffer + 8, maxBytes - 4)
      lenView.setUint32(4,numBytes,true);
      console.log(`read wrote ${lenView.getUint32(4,true)} bytes, remainder: ${stdin}`)
      console.log(`read blen:${lenView.getUint32(0,true)} cstrlen:${lenView.getUint32(4,true)}`)   },
    sys_listen: (reqAddr, cbAddr) => {
      //TODO: call into the module to wasm fn "event" to push an event
      //to forth, which pushes esi to the return stack and queues the
      //callback function provided from listen. reqaddr could be
      //"fetch", in which case no channel is used.  otherwise very
      //similar to sys_fetch.
    },
    sys_fetch: (channel, reqAddr) => {
      //TODO: map to fetch promise, write to channel buffer,
      //javascript "fetch" as fallback, explicit handles for
      //"textEntry" or any third party protocols like activitypub
      console.log(`fetch ${channel} ${reqAddr}`)
    },
    sys_echo: (val) => output.print(`${val} `),
    sys_echochar: (val) => output.print(String.fromCharCode(val)),
    sys_reflect: (addr) => {
      console.log(`reflect: ${addr}: ${
        new DataView(wasmMem.buffer, addr, 4)
              .getUint32(0,true)
      }`)
    },
    vocab_get: (addr) => {
      const bytes = new DataView(
	wasmMem.buffer,
	addr,
	4
      ).getUint32(0,true)
      const word = String.fromCharCode.apply(
	null,
	new Uint16Array(wasmMem.buffer, addr + 4, bytes >> 1)
      )
      const answer = dictionary[word.toUpperCase()]
      console.log(`vocab_get ${word}: ${answer}`)
      if (answer === undefined)
	return 0
      return answer
    },
    vocab_set: (addr, num) => {
      const bytes = new DataView(
	wasmMem.buffer,
	addr,
	4
      ).getUint32(0,true)
      const word = String.fromCharCode.apply(
	null,
	new Uint16Array(wasmMem.buffer, addr + 4, bytes >> 1)
      )
      console.log(`vocab_set ${word}: ${num}`)
      dictionary[word.toUpperCase()] = num
      return 0
    },
    is_whitespace: (key) => /\s/.test(String.fromCharCode(key)),
    sys_stack: () => console.log(`[${simstack}]`),
    sys_parsenum: (addr, base) => {
      const byteV = new DataView(
	wasmMem.buffer,
	addr,
	4
      )
      const word = String.fromCharCode.apply(
	null,
	new Uint16Array(wasmMem.buffer, addr + 4, byteV.getUint32(0,true) >> 1)
      )
      const answer = Number.parseInt(word, base)
      byteV.setUint32(0,Number.isNaN(answer),true)
      return answer
    },
    sys_words: () => {
      output.print(Object.getOwnPropertyNames(dictionary).toString().split(',').join('  '))
    }
  }
}

fetch('forth.wasm')
  .then(re => re.arrayBuffer())
  .then(buf => WebAssembly.instantiate(buf, wasmImport))
  .then(result => {
    wasmMem = result.instance.exports.memory
    forth = result.instance.exports.main
    console.log('wasm loaded');
    forth()
  })
